@@ -84,11 +84,43 @@ string getPathOfReleaseList(const Config& config, const IndexEntry& entry)
 	return getPathOfIndexEntry(config, entry) + "_Release";
 }
 
+string getPathOfInReleaseList(const Config& config, const IndexEntry& entry)
+{
+	return getPathOfIndexEntry(config, entry) + "_InRelease";
+}
+
+namespace {
+
+string selectNewerFile(const string& leftPath, const string& rightPath)
+{
+	bool leftExists = fs::fileExists(leftPath);
+	bool rightExists = fs::fileExists(rightPath);
+	if (!leftExists && !rightExists) return string();
+	if (!rightExists) return leftPath;
+	if (!leftExists) return rightPath;
+	return fs::fileModificationTime(leftPath) >= fs::fileModificationTime(rightPath) ?
+			leftPath : rightPath;
+}
+
+}
+
+string getPathOfMasterReleaseLikeList(const Config& config, const IndexEntry& entry)
+{
+	return selectNewerFile(
+			getPathOfInReleaseList(config, entry),
+			getPathOfReleaseList(config, entry));
+}
+
 string getDownloadUriOfReleaseList(const IndexEntry& entry)
 {
 	return getUriOfIndexEntry(entry) + "/Release";
 }
 
+string getDownloadUriOfInReleaseList(const IndexEntry& entry)
+{
+	return getUriOfIndexEntry(entry) + "/InRelease";
+}
+
 static string getIndexListSuffix(const Config& config, const IndexEntry& entry, char delimiter)
 {
 	if (entry.component.empty())
@@ -121,24 +153,19 @@ string getPathOfIndexList(const Config& config, const IndexEntry& entry)
 	return basePath + "_" + indexListSuffix;
 }
 
-static vector< Cache::IndexDownloadRecord > getDownloadInfoFromRelease(
+static vector< FileDownloadRecord > getDownloadInfoFromRelease(
 		const Config& config, const IndexEntry& indexEntry, const string& suffix)
 {
 	auto baseUri = getUriOfIndexEntry(indexEntry);
 	// TODO: make cachefiles::getAlias* functions
 	auto alias = indexEntry.uri + ' ' + indexEntry.distribution;
 
-	vector< Cache::IndexDownloadRecord > result;
+	vector< FileDownloadRecord > result;
 
 	try
 	{
-		string openError;
-		auto releaseFilePath = getPathOfReleaseList(config, indexEntry);
-		File releaseFile(releaseFilePath, "r", openError);
-		if (!openError.empty())
-		{
-			fatal2(__("unable to open the file '%s': %s"), releaseFilePath, openError);
-		}
+		auto releaseFilePath = getPathOfMasterReleaseLikeList(config, indexEntry);
+		RequiredFile releaseFile(releaseFilePath, "r");
 
 		HashSums::Type currentHashSumType = HashSums::Count;
 		// now we need to find if this variant is present in the release file
@@ -190,8 +217,8 @@ static vector< Cache::IndexDownloadRecord > getDownloadInfoFromRelease(
 				}
 				if (!foundRecord)
 				{
-					Cache::IndexDownloadRecord& record =
-							(result.push_back(Cache::IndexDownloadRecord()), *(result.rbegin()));
+					FileDownloadRecord& record =
+							(result.push_back(FileDownloadRecord()), *(result.rbegin()));
 					record.uri = uri;
 					record.size = string2uint32(m[2]);
 					record.hashSums[currentHashSumType] = m[1];
@@ -215,7 +242,7 @@ static vector< Cache::IndexDownloadRecord > getDownloadInfoFromRelease(
 	return result;
 }
 
-vector< Cache::IndexDownloadRecord > getDownloadInfoOfIndexList(
+vector< FileDownloadRecord > getDownloadInfoOfIndexList(
 		const Config& config, const IndexEntry& indexEntry)
 {
 	return getDownloadInfoFromRelease(config, indexEntry,
@@ -308,27 +335,6 @@ vector< pair< string, string > > getPathsOfLocalizedDescriptions(
 	return result;
 }
 
-vector< Cache::LocalizationDownloadRecord > getDownloadInfoOfLocalizedDescriptions(
-		const Config& config, const IndexEntry& entry)
-{
-	auto chunkArrays = getChunksOfLocalizedDescriptions(config, entry);
-	auto basePath = getPathOfIndexEntry(config, entry);
-	auto baseUri = getUriOfIndexEntry(entry);
-
-	vector< Cache::LocalizationDownloadRecord > result;
-
-	FORIT(chunkArrayIt, chunkArrays)
-	{
-		Cache::LocalizationDownloadRecord record;
-		record.localPath = basePath + "_" + join("_", *chunkArrayIt);
-		// yes, somewhy translations are always bzip2'ed
-		record.uri = baseUri + "/" + join("/", *chunkArrayIt) + ".bz2";
-		result.push_back(std::move(record));
-	}
-
-	return result;
-}
-
 vector< FileDownloadRecord > getDownloadInfoOfLocalizationIndex(const Config& config,
 		const IndexEntry& entry)
 {
@@ -385,6 +391,36 @@ string getPathOfExtendedStates(const Config& config)
 	return config.getPath("dir::state::extendedstates");
 }
 
+namespace {
+
+bool openingForReadingSucceeds(const string& path, const string& fileType, bool debugging)
+{
+	string openError;
+	File file(path, "r", openError);
+	if (!openError.empty())
+	{
+		if (debugging)
+		{
+			debug2("unable to read %s file '%s': %s", fileType, path, openError);
+		}
+		return false;
+	}
+	return true;
+}
+
+bool gpgLineShouldBeSkipped(const string& line)
+{
+	static const vector<string> ignoreTypes = { "SIG_ID ", "NEWSIG", "KEY_CONSIDERED " };
+	for (const auto& type: ignoreTypes)
+	{
+		if (line.find(type) != string::npos)
+			return true;
+	}
+	return false;
+}
+
+}
+
 bool verifySignature(const Config& config, const string& path, const string& alias)
 {
 	auto debugging = config.getBool("debug::gpgv");
@@ -394,10 +430,8 @@ bool verifySignature(const Config& config, const string& path, const string& ali
 	}
 
 	auto keyringPath = config.getString("gpgv::trustedkeyring");
-	if (debugging)
-	{
-		debug2("keyring file is '%s'", keyringPath);
-	}
+	if (debugging) debug2("keyring file is '%s'", keyringPath);
+	if (!openingForReadingSucceeds(keyringPath, "keyring", debugging)) return false;
 
 	auto signaturePath = path + ".gpg";
 	if (debugging)
@@ -407,38 +441,14 @@ bool verifySignature(const Config& config, const string& path, const string& ali
 
 	if (!fs::fileExists(signaturePath))
 	{
-		if (debugging)
-		{
-			debug2("signature file '%s' doesn't exist", signaturePath);
-		}
-		return 0;
+		if (debugging) debug2("signature file '%s' doesn't exist, omitting it and assuming self-signed file", signaturePath);
+		signaturePath.clear();
+	}
+	else
+	{
+		if (!openingForReadingSucceeds(signaturePath, "signature", debugging)) return false;
 	}
 
-	// file checks
-	{
-		string openError;
-		File file(signaturePath, "r", openError);
-		if (!openError.empty())
-		{
-			if (debugging)
-			{
-				debug2("unable to read signature file '%s': %s", signaturePath, openError);
-			}
-			return false;
-		}
-	}
-	{
-		string openError;
-		File file(keyringPath, "r", openError);
-		if (!openError.empty())
-		{
-			if (debugging)
-			{
-				debug2("unable to read keyring file '%s': %s", keyringPath, openError);
-			}
-			return false;
-		}
-	}
 
 	bool verifyResult = false;
 	try
@@ -455,31 +465,19 @@ bool verifySignature(const Config& config, const string& path, const string& ali
 		smatch m;
 		auto gpgGetLine = [&gpgPipe, &m, &debugging]() -> string
 		{
-			static const sregex sigIdRegex = sregex::compile("\\[GNUPG:\\] SIG_ID");
-			static const sregex generalRegex = sregex::compile("\\[GNUPG:\\] ");
+			static const sregex gnupgPrefixRegex = sregex::compile("\\[GNUPG:\\] ");
 			string result;
+
 			do
 			{
 				gpgPipe.getLine(result);
-				if (debugging && !gpgPipe.eof())
-				{
-					debug2("fetched '%s' from gpg pipe", result);
-				}
-			} while (!gpgPipe.eof() && (
-						regex_search(result, m, sigIdRegex, regex_constants::match_continuous) ||
-						!regex_search(result, m, generalRegex, regex_constants::match_continuous)));
+				if (debugging && !gpgPipe.eof()) debug2("fetched '%s' from gpg pipe", result);
+				result = regex_replace(result, gnupgPrefixRegex, "");
+			} while (!gpgPipe.eof() && gpgLineShouldBeSkipped(result));
 
-			if (gpgPipe.eof())
-			{
-				return "";
-			}
-			else
-			{
-				return regex_replace(result, generalRegex, "");
-			}
+			return gpgPipe.eof() ? "" : result;
 		};
 
-
 		auto status = gpgGetLine();
 		if (status.empty())
 		{
@@ -499,35 +497,19 @@ bool verifySignature(const Config& config, const string& path, const string& ali
 
 		if (messageType == "GOODSIG")
 		{
-			string furtherInfo = gpgGetLine();
-			if (!regex_match(furtherInfo, m, messageRegex))
-			{
-				fatal2(__("gpg: '%s': invalid detailed information string '%s'"), alias, furtherInfo);
-			}
-
-			string furtherInfoType = m[1];
-			string furtherInfoMessage = m[2];
-			if (furtherInfoType == "VALIDSIG")
-			{
-				// no comments :)
-				verifyResult = 1;
-			}
-			else if (furtherInfoType == "EXPSIG")
-			{
-				warn2(__("gpg: '%s': expired signature: %s"), alias, furtherInfoMessage);
-			}
-			else if (furtherInfoType == "EXPKEYSIG")
-			{
-				warn2(__("gpg: '%s': expired key: %s"), alias, furtherInfoMessage);
-			}
-			else if (furtherInfoType == "REVKEYSIG")
-			{
-				warn2(__("gpg: '%s': revoked key: %s"), alias, furtherInfoMessage);
-			}
-			else
-			{
-				warn2(__("gpg: '%s': unknown error: %s %s"), alias, furtherInfoType, furtherInfoMessage);
-			}
+			verifyResult = 1;
+		}
+		else if (messageType == "EXPSIG")
+		{
+			warn2(__("gpg: '%s': expired signature: %s"), alias, message);
+		}
+		else if (messageType == "EXPKEYSIG")
+		{
+			warn2(__("gpg: '%s': expired key: %s"), alias, message);
+		}
+		else if (messageType == "REVKEYSIG")
+		{
+			warn2(__("gpg: '%s': revoked key: %s"), alias, message);
 		}
 		else if (messageType == "BADSIG")
 		{
@@ -536,44 +518,30 @@ bool verifySignature(const Config& config, const string& path, const string& ali
 		else if (messageType == "ERRSIG")
 		{
 			// gpg was not able to verify signature
-
-			// maybe, public key was not found?
-			bool publicKeyWasNotFound = false;
-			auto detail = gpgGetLine();
-			if (!detail.empty())
+			auto parts = split(' ', message); // <keyid> <pkalgo> <hashalgo> <sig_class> <time> <rc>
+			if (parts.size() != 6)
 			{
-				if (!regex_match(detail, m, messageRegex))
-				{
-					fatal2(__("gpg: '%s': invalid detailed information string '%s'"), alias, detail);
-				}
-				string detailType = m[1];
-				string detailMessage = m[2];
-				if (detailType == "NO_PUBKEY")
-				{
-					publicKeyWasNotFound = true;
-
-					// the message looks like
-					//
-					// NO_PUBKEY D4F5CE00FA0E9B9D
-					//
-					warn2(__("gpg: '%s': public key '%s' is not found"), alias, detailMessage);
-				}
+				fatal2(__("gpg: '%s': invalid detailed information string '%s'"), alias, message);
 			}
+			const auto& rc = parts.back();
 
-			if (!publicKeyWasNotFound)
+			if (rc == "9")
+			{
+				warn2(__("gpg: '%s': public key '%s' is not found"), alias, parts[0]);
+			}
+			else if (rc == "4")
+			{
+				warn2(__("gpg: '%s': unknown algorithm '%s/%s'"), alias, parts[1], parts[2]);
+			}
+			else
 			{
 				warn2(__("gpg: '%s': could not verify a signature: %s"), alias, message);
 			}
 		}
 		else if (messageType == "NODATA")
 		{
-			// no signature
 			warn2(__("gpg: '%s': empty signature"), alias);
 		}
-		else if (messageType == "KEYEXPIRED")
-		{
-			warn2(__("gpg: '%s': expired key: %s"), alias, message);
-		}
 		else
 		{
 			warn2(__("gpg: '%s': unknown message: %s %s"), alias, messageType, message);
@@ -602,20 +570,15 @@ shared_ptr< cache::ReleaseInfo > getReleaseInfo(const Config& config,
 	smatch matches;
 	try
 	{
-		string openError;
-		File file(path, "r", openError);
-		if (!openError.empty())
-		{
-			fatal2(__("unable to open the file '%s': %s"), path, openError);
-		}
+		RequiredFile file(path, "r");
 
 		string line;
 		while (! file.getLine(line).eof())
 		{
-			if (!regex_match(line, matches, fieldRegex))
-			{
-				break;
-			}
+			if (line.empty()) continue;
+			if (line[0] == '-') continue; // "----- BEGIN PGP SIGNED MESSAGE-----"
+			if (!regex_match(line, matches, fieldRegex)) break;
+
 			string fieldName = matches[1];
 			string fieldValue = matches[2];
 
